#include <OneWire.h>
#include <DallasTemperature.h>
#include <SimpleTimer.h>
#include <CapacitiveSensor.h>
#include <SoftwareServo.h>    // Important! Servo judders with Servo.h
#include "setPWMPrescaler.h"; // Custom .h file allows the adjustment of the PWM frequencies on the AtMega 32u4

//
// ==================================
// VERSION COMMENTS
// ==================================
//
// V.0.9 Jan. 2. 2015 MacTester (Working with NUC-Dummy inside iMac G5)
// V.0.9.1 enhanced PSU shutdown delay from 20 to 30s and reduced PS impulse from 1 to 0.3s in order to solve weird hang issues
// V.0.9.2 fixed inverter value mapping to 0-100, 140 - 0, PSU ts = millis(), if PS pressed...
// V.0.9.4 State detection: S5, if LOW, LOW. Inverter startup brightness 70%. PS impulse 0.7s -> Runs OK!
// V.0.9.5 state = 1, if power switch pressed
// V.0.9.6 Touch sensors 1M resistors instead of 10M, Fan PWM frequency reduced from 31.5kHz to 123Hz (original iMac =100Hz)
// V.0.9.6_Tach_new_Pinout for testing kiwis fan interrupts. Caution, different pinout!

//
// ==================================
// PIN ASSIGNMENTS & GLOBAL VARIABLES
// ==================================
//

//
// Digital 0, 1, Shared as serial pins (Serial1)
// Digital 2 - 13, General Purpose
// Digital 3, 5, 6, 9, 10, 11, and 13 - PWM
// Digital 11 - 13 Excluded on Pro Micro
// Digital 13 has (not Leo) an LED and resistor attached to it
// Digital 14 - 16, Shared with an SPI bus
// Digital 17 - Doubles and RX LED
// Digital Pins can sink 40ma, recommended 470 ohm resistor
// Analog Inputs: A0-A3, Pro Micro excludes A4, A5
// Analog Inputs: A6-A11 (on digital pins 4, 6, 8, 9, 10, and 12)
// Analog Pins can be used as digital ( refer to them as A0, A1, ...)
//

//----PSU CONTROLLER
const byte PSU_DEACTIVATE_POUT = 4; // if HIGH, the main PSU will be OFF -> (only 5V Standby is available)

//----FAN CONTROLLER

// Temperature Sensor Pin & other parameters
#define ONE_WIRE_BUS A2 //One Wire BUS is Pin 16 (all Temperature Sensors are wired there)
OneWire oneWire(ONE_WIRE_BUS); //Setup a oneWire instance to communicate with OneWire Devices
DallasTemperature sensors(&oneWire); //Pass our oneWire reference to Dallas Temperature
DeviceAddress tempDeviceAddress0;
DeviceAddress tempDeviceAddress1;
DeviceAddress tempDeviceAddress2;
int  resolution = 10; //Sensor resolution = 10bit = 0.25°C increment (9 to 12bit possible)

//Fan PWM Output
const byte FAN1_CONTROL_POUT = 6; // PWM Fan Speed Output (HDD Fan)
const byte FAN2_CONTROL_POUT = 10; // PWM Fan Speed Output (CPU Fan)
const byte FAN3_CONTROL_POUT = 9; // PWM Fan Speed Output (Case Fan)

//Fan Tach Inputs  (with 3.3yf capacitor in parallel, between Ax Pin and GND)

const byte FAN1_TACH_INT = 1; // Fan1 Speed Tachometer Input
const byte FAN2_TACH_INT = 2; // Fan2 Speed Tachometer Input
const byte FAN3_TACH_INT = 7; // Fan3 Speed Tachometer Input

volatile long fan1RotationCount;
volatile unsigned long fan1TimeStart;
volatile long fan2RotationCount;
volatile unsigned long fan2TimeStart;
volatile long fan3RotationCount;
volatile unsigned long fan3TimeStart;

//----LED CONTROLLER, POWER SWITCH & NUC CS HEADER
const int SLEEP_LED_POUT = 3;    // LED connected to digital pin 3
const int CASE_POWER_SWITCH_PIN = A1;  // Case Switch
const int NUC_POWER_LED_CS_PIN2 = 14;  // NUC Custom Solution Header Pin 2
const int NUC_POWER_LED_CS_PIN4 = 15;  // NUC Custom Solution Header Pin 4

//----LCD INVERTER CONTROLLER
const int INVERTER_POUT = 5;    // CCFL inverter

//----SERVO CONTROLLER
const int DOOR_SERVO_POUT = 0;    // Door servo connected to digital pin 0 RXI (a PWM pin is not required for servos)
SoftwareServo doorServo;

//----CAPACITIVE TOUCH SENSORS
// Up and down capacitive touch brightness pins
const byte TOUCH_COMMON_PIN_CAP = A0;
const byte TOUCH_DOWN_PIN_CAP = 16;
const byte TOUCH_UP_PIN_CAP = 8;

// 1M resistor between pins 2 & 8, pin 8 is sensor pin
CapacitiveSensor dnSensor = CapacitiveSensor(TOUCH_COMMON_PIN_CAP, TOUCH_DOWN_PIN_CAP);

// 1M resistor between pins 2 & 7, pin 7 is sensor pin
CapacitiveSensor upSensor = CapacitiveSensor(TOUCH_COMMON_PIN_CAP, TOUCH_UP_PIN_CAP);

//----APPLE CHIME START MELODY
const byte CHIME_POUT = A3;

//----TIMER OBJECT
SimpleTimer timer;

//----DEBUG
boolean debug = false; // Set to "false" for operation! Otherwise the Brightness Menulet will not work properly.

//
// ========================================
// MAIN ARDUINO SETUP (1x during startup)
// ========================================
//

void setup() {
  //Start Serial Port
  //Serial.begin(9600);

  // Start up the One Wire Sensor Library
  sensors.begin();
  sensors.getAddress(tempDeviceAddress0, 0);
  sensors.getAddress(tempDeviceAddress1, 1);
  sensors.getAddress(tempDeviceAddress2, 2);
  sensors.setResolution(tempDeviceAddress0, resolution);
  sensors.setResolution(tempDeviceAddress1, resolution);
  sensors.setResolution(tempDeviceAddress2, resolution);

  //Set PinModes

  pinMode(PSU_DEACTIVATE_POUT, OUTPUT);
  digitalWrite(PSU_DEACTIVATE_POUT, LOW); //PSU on after mains was switched on!
 
  pinMode(FAN1_CONTROL_POUT, OUTPUT);
  pinMode(FAN2_CONTROL_POUT, OUTPUT);
  pinMode(FAN3_CONTROL_POUT, OUTPUT);

  pinMode(FAN1_TACH_INT, INPUT_PULLUP);
  pinMode(FAN2_TACH_INT, INPUT_PULLUP);
  pinMode(FAN3_TACH_INT, INPUT_PULLUP);

  pinMode(SLEEP_LED_POUT, OUTPUT);
  pinMode(CASE_POWER_SWITCH_PIN, INPUT_PULLUP);
  pinMode(NUC_POWER_LED_CS_PIN2, INPUT_PULLUP);
  pinMode(NUC_POWER_LED_CS_PIN4, INPUT_PULLUP);

  pinMode(CHIME_POUT, OUTPUT);
  digitalWrite(CHIME_POUT, LOW);

  pinMode(INVERTER_POUT, OUTPUT);

  //Attach servo to Pin
  doorServo.attach(DOOR_SERVO_POUT);

  // Timer, which triggers the sleep LED brightness attenuator every 10ms
  timer.setInterval(10, ledAttenuator);

  // Timer, which triggers the touch sensor reading every 182ms
  timer.setInterval(182, touchControl);
  
  // Timer, which triggers the fan reading printout reading every 2s
  timer.setInterval(2000, printFanDetails);

  // Adjust PWM frequencies
  // Note: the original iMac fans are driven @ 100Hz only!
  setPWMPrescaler(FAN1_CONTROL_POUT, 256); // Sets 123Hz Frequency ( 256 )
  setPWMPrescaler(FAN2_CONTROL_POUT, 256); // Sets 123Hz Frequency ( 256 )
  setPWMPrescaler(FAN3_CONTROL_POUT, 256); // Sets 123Hz Frequency ( 256 )

  setPWMPrescaler(INVERTER_POUT, 1); // Sets 31.25KHz Frequency ( 1 )
  
  // Fan tachometer initialisation   
  fan1RotationCount = 0;
  fan2RotationCount = 0;
  fan3RotationCount = 0;
  fan1TimeStart = millis();
  fan2TimeStart = millis();
  fan3TimeStart = millis();
  attachInterrupt(3,interrupFan1,FALLING); // pin 1 (TX0) = interrupt 3
  attachInterrupt(1,interrupFan2,RISING); // pin 2 = interrupt 1
  attachInterrupt(4,interrupFan3,FALLING); // pin 7 = interrupt 4

  //Calibrate touch sensors
  delay (500);                // Wait 0.5s...
  setupTouch();               // ... then initialise touch sensors
  
  // Inverter setup
  setupInverter();
}

//
// ========================================
// MAIN LOOP
// ========================================
//

void loop() {

  // Loop the timer
  timer.run();

  //Loop the Power Switch FSM
  fsmPowerSwitch();

  //Loop the PSU Controller FSM
  fsmPSU();

  //Loop the Fan Controller FSM
  fsmFans();

  //Loop the LED Controller
  ledSinus();

  //Loop servo function & refresh
  servoFunction();
  SoftwareServo::refresh();

  //Loop USB serial controller
  loopSerialCommandProcessor();

  //Loop the Chime FSM
  fsmChime();
}

//
// =========================================
// NUC POWER STATE DETECTION
// =========================================
//

byte powerStateRawValue() {

  // read voltages on two CS header pins
  int POS_Val = digitalRead(NUC_POWER_LED_CS_PIN2);
  int Neg_Val = digitalRead(NUC_POWER_LED_CS_PIN4);

  // the state we will return
  static byte value = 0;

  // if the PIN Voltages are correct polarity set S3 State
  if ( POS_Val == HIGH && Neg_Val == LOW  ) value = 0;  // S1 = All ON
  if ( POS_Val == LOW  && Neg_Val == HIGH ) value = 3;  // S3 = Standby Mode
  if ( POS_Val == HIGH && Neg_Val == HIGH ) value = 5;  // S5 = Soft Off (only 5V standby is ON, all other supply voltages are OFF)
  if ( POS_Val == LOW && Neg_Val == LOW ) value = 5;

  // return that state
  return value;
}

byte powerState() {

  // variable used between calls
  static byte reportedState = 0; // value we last returned
  static byte lastObservered = 0; // value we last observed
  static unsigned long firstObservervedWhen = 0; // when did we first observe

  // read the raw state from the input pins (see above)
  byte justObservered = powerStateRawValue();

  // if the just observered state is different to last observered then
  if ( justObservered != lastObservered ) {

    // then we want to reset our observation
    lastObservered = justObservered;

    // and reset a timer so know when we first observered the new value
    firstObservervedWhen = millis();
  }

  // if the last observered value is different to last reported
  // and sufficient time has elapsed since we first observed the new value
  if ( lastObservered != reportedState && firstObservervedWhen - millis() > 100 ) {

    // then we can set the last observered value is probably final so
    reportedState = lastObservered;
  }

  return reportedState;
}

//
// =========================================
// POWER SWITCH FSM (Finite State Machine)
// =========================================
//

void fsmPowerSwitch() {

  static int state = 1; // initial state is 1, the "idle" state.
  static unsigned long ts;  // To store the "current" time in for delays.
  int switchVal = digitalRead(CASE_POWER_SWITCH_PIN);  //Read switch

  switch (state)
  {

      //----Is Power Switch pressed? -> Generate a 0.7s "LOW-Pulse", even if the power swich button was released before!
      //    Reason: The PSU must be enabled before the NUC can be switched on!
    case 1:
      if (switchVal == false) {
        pinMode(CASE_POWER_SWITCH_PIN, OUTPUT);  //Make it an Output...
        digitalWrite(CASE_POWER_SWITCH_PIN, LOW);  //And pull it LOW for 0.7s
        state = 2;
      }
      else state = 4;  // Else don't make Pin an output
      break;

      //----Start timer
    case 2:
      ts = millis();  // Remember the current time
      state = 3;
      break;

      //----If 0.5s has passed...
    case 3:
      if (millis() > ts + 700)
      {
        state = 4;
      }
      break;

      //----Make power switch Pin an input again
    case 4:
      pinMode(CASE_POWER_SWITCH_PIN, INPUT_PULLUP);
      state = 1;
      break;

      //----
    default:
      state = 1;
      break;
  }
}


//
// =========================================
// PSU CONTROLLER FSM
// =========================================
//

void fsmPSU() {

  static int state = 1; // initial state is 1, the "idle" state.
  static unsigned long ts;  // To store the "current" time in for delays.
  int switchVal = digitalRead(CASE_POWER_SWITCH_PIN);  //Read switch

  // Always enable the main PSU, if the power switch is pressed!
  if (switchVal == false) { //false = Switch pressed!
    digitalWrite(PSU_DEACTIVATE_POUT, LOW); // LOW = PSU enabled
    ts = millis();  // Remember the current time
    state = 1;
  }

  switch (state)
  {

      //----Initial state
    case 1:
      // We don't need to do anything here, waiting for a forced state change.
      state = 2;
      break;

      //----Is Computer in S5 (OFF)?
    case 2:
      if (powerState() == 5) state = 3; // wait 30s, then check it again
      else state = 7;  // Else don't deactivate PSU!!
      break;

      //----Start timer
    case 3:
      ts = millis();  // Remember the current time
      state = 4;
      break;

      //----If time has passed, check it again for safety...
    case 4:
      if (millis() > ts + 30000)
      {
        state = 5;
      }
      break;

      //----Check it again: Is Computer in S5 (OFF)?
    case 5:
      if (powerState() == 5) state = 6; // deactivate PSU!!
      else state = 7;  // Else don't deactivate PSU!!
      break;

      //----!!!Deactivate Main PSU!!!
    case 6:
      digitalWrite(PSU_DEACTIVATE_POUT, HIGH); // HIGH = PSU OFF!!
      state = 7;
      break;

      //----Start timer
    case 7:
      ts = millis();  // Remember the current time
      state = 8;
      break;

      //----If time has passed, go back to state 1
    case 8:
      if (millis() > ts + 100)
      {
        state = 1;
      }
      break;

      //----
    default:
      state = 1;
      break;
  }
}

//
// =========================================
// APPLE CHIME START SOUND FSM
// =========================================
//

void fsmChime() {

  static int state = 1; // initial state is 1, the "idle" state.
  static unsigned long ts;  // To store the "current" time in for delays.
  static boolean wasInS5 = false;

  // if S3 State (Standby Mode) is active, reset bit "wasInS5"
  if (powerState() == 3) {
    wasInS5 = false;
  }

  // if S5 State (OFF Mode) is active, set bit "wasInS5"
  if (powerState() == 5) {
    wasInS5 = true;
  }

  switch (state)
  {

      //----Initial state
    case 1:
      // We don't need to do anything here, waiting for a forced state change.
      state = 2;
      break;

      //----Was Computer in S5 and is in S0 now?
    case 2:
      if ( wasInS5 == true && powerState() == 0) // LED ON, delay 5s, then trigger Apple Chime
      {
        digitalWrite(SLEEP_LED_POUT, HIGH);
        state = 3;
      }
      else state = 6;  // Don't play Chime
      break;

      //----Start timer
    case 3:
      ts = millis();  // Remember the current time
      state = 4;
      break;

      //----If time has passed, trigger Chime
    case 4:
      if (millis() > ts + 5000) // delay 5s
      {
        state = 5;
      }
      break;

      //----LED OFF, Trigger Chime, reset bit "wasInS5"
    case 5:
      digitalWrite(CHIME_POUT, HIGH);
      digitalWrite(SLEEP_LED_POUT, LOW);
      wasInS5 = false;
      state = 6;
      break;

      //----Start timer
    case 6:
      ts = millis();  // Remember the current time
      state = 7;
      break;

      //----If time has passed, go back to state 1
    case 7:
      if (millis() > ts + 2700) // Keep output for at least the duration of the Apple Chime high to trigger speaker switch relay!
      {
        digitalWrite(CHIME_POUT, LOW);
        state = 1;
      }
      break;

      //----
    default:
      state = 1;
      break;
  }
}

//
// =========================================
// TOUCH SENSORS
// =========================================
//

void setupTouch() {
  void reset_CS_AutoCal();    // calibrate the touch sensors
  //upSensor.set_CS_AutocaL_Millis(0xBB8);     // autocalibrate timeout 3000ms  = HEX (0xBB8)
  //dnSensor.set_CS_AutocaL_Millis(0xBB8);
  }

const int arrayLength = 10;  // declares, how much samples are used in the smoothing function

// Varialbles and references for upSensor smoothing
int upReadings[arrayLength];
int upPointer = 0;              // the pointer of the current reading
int& upPointerRef = upPointer;  // the reference for passing the pointer to the function and back!
long upTotal = 0;               // the total
long& upTotalRef = upTotal;     // the reference for passing the total to the function and back!

// Varialbles and references for dnSensor smoothing
int dnReadings[arrayLength];
int dnPointer = 0;              // the pointer of the current reading
int& dnPointerRef = dnPointer;  // the reference for passing the pointer to the function and back!
long dnTotal = 0;               // the total
long& dnTotalRef = dnTotal;     // the reference for passing the total to the function and back!

// Main Touch Sensor Reading
void touchControl() {

  //Read both sensors
  long start = millis();
  int upReading = upSensor.capacitiveSensor(80); // (XX) is the number of sensor samples
  int dnReading = dnSensor.capacitiveSensor(80);

  // Smooth Sensor Readings (average of x = "arrayLength" samples)
  //upReading = smoothInt(upReading, upReadings, arrayLength, upPointerRef, upTotalRef);
  //dnReading = smoothInt(dnReading, dnReadings, arrayLength, dnPointerRef, dnTotalRef);


  if (powerState() == 0) { // Only allow brightness adjustment, if Computer is in (S0 State)

    // If the capacitive sensor reads above a certain threshold,
    if (upReading > 70 ) {  // 200
      increaseBrightness();
      if (debug) {
        Serial.println("Brightness increased");
      }
    }
    if (dnReading > 70 ) {
      decreaseBrightness();
      if (debug) {
        Serial.println("Brightness decreased");
      }
    }
  }

  if (debug) {
    Serial.print(millis() - start);  // check on performance in ms > 25 causes servo judder! (depends on resistor & sensor samples above)
    Serial.print("\t");
    Serial.print(dnReading);
    Serial.print("\t");
    Serial.println(upReading);
  }
}

//
// =========================================
// FAN CONTROLLER FSM
// =========================================
//

void fsmFans() {

  static int state = 1; // initial state is 1, the "idle" state.
  static unsigned long ts;  // To store the "current" time in for delays.

  switch (state)
  {

      //----Initial state
    case 1:
      // We don't need to do anything here, waiting for a forced state change.
      state = 2;
      break;

      //----Wait
    case 2:
      //Do nothing
      ts = millis();  // Remember the current time
      state = 3;  // Move to the next state
      break;

      //----Wait for time
    case 3:
      // If time has passed, then move on to the next state.
      if (millis() > ts + 1200) {
        state = 4;
      }
      break;

      //----Request sensors for readings
    case 4:
      sensors.setWaitForConversion(false);  // makes it async
      sensors.requestTemperatures(); //Request temperatures (from all connected sensors)
      sensors.setWaitForConversion(true);
      ts = millis();  // Remember the current time
      state = 5;
      break;

      //----Wait for sensors (in accordance with sensor datasheet, required time depends on bit resolution!)
    case 5:
      // If time has passed, then go to state 6.
      if (millis() > ts + 800)
      {
        state = 6;
      }
      break;

      //----Read sensors, write fan PWM
    case 6:
      // Read sensor 0, write fan 1 PWM (HDD Fan)
      readTemperatureSensor(0, FAN1_CONTROL_POUT, FAN1_TACH_INT, 70, 255, 1000); //Sensor No, Fan No, Tach No, minPWM (15), maxPWM, Divider235
      // Read sensor 1, write fan 2 PWM (CPU Fan)
      readTemperatureSensor(1, FAN2_CONTROL_POUT, FAN2_TACH_INT, 70, 255, 1000); //Sensor No, Fan No, Tach No, minPWM (13), maxPWM, Divider
      // Read sensor 2, write fan 3 PWM (Case Fan)
      readTemperatureSensor(2, FAN3_CONTROL_POUT, FAN3_TACH_INT, 70, 255, 500); //Sensor No, Fan No, Tach No, minPWM (120), maxPWM, Divider (500)
      state = 1;
      break;

      //----
    default:
      state = 1;
      break;
  }
}

//Read sensor, detect fan rotation & write fan PWM function (called from above)
void readTemperatureSensor(int sensorNo, int fanNo, int tachNo, int pwmMin, int pwmMax, int divider) {
  float sensorReading = (sensors.getTempCByIndex(sensorNo)); //Read temperature from sensor
  float pwmValue = sensorReading * sensorReading * sensorReading / divider;

  if (analogRead(tachNo) > 600) {  //does fan spin? (tach > 600 means no)
    pwmMin = (pwmMin + pwmMax) / 2;  //Allows fan to start!
  }

  pwmValue = constrain(pwmValue, pwmMin, pwmMax);  //Limit PWM range

  if (powerState() != 0) {
    pwmValue = 0;  // Fan off, if computer is not in S0 (ON)
  }
  analogWrite(fanNo, pwmValue); //Write Fan PWM

  if (debug) {
    Serial.print("Sensor ");
    Serial.print(sensorNo);
    Serial.print(" ");
    Serial.print(sensorReading);
    Serial.print(" Celsius  =  Fan ");
    Serial.print(fanNo);
    Serial.print(" ");
    Serial.print(pwmValue);
    Serial.println(" PWM");
  }
}

// Interrupt functions for fan tach
void interrupFan1() {
  fan1RotationCount ++;
}

void interrupFan2() {
  fan2RotationCount ++;
}

void interrupFan3() {
  fan3RotationCount ++;
}

// Function for printing the fan RPM
void printFanDetails() {
  if (debug) {
    Serial.print("RPM");
    Serial.print(F("\t"));
    Serial.print(getFanRPM1());
    Serial.print(F("\t"));
    Serial.print(getFanRPM2());
    Serial.print(F("\t"));
    Serial.print(getFanRPM3());
    Serial.println();
  }
}

int getFanRPM1() {
  double ret = fan1RotationCount /2L *1000L * 60L / ( millis() - fan1TimeStart );
  fan1RotationCount = 0;
  fan1TimeStart = millis();
  return  ret;
}

int getFanRPM2() {
  double ret = fan2RotationCount /2L * 1000L * 60L / ( millis() - fan2TimeStart );
  fan2RotationCount = 0;
  fan2TimeStart = millis();
  return  ret;
}

int getFanRPM3() {
  double ret = fan3RotationCount /2L * 1000L * 60L / ( millis() - fan3TimeStart );
  fan3RotationCount = 0;
  fan3TimeStart = millis();
  return  ret;
}

//
// ========================================
// SLEEP LED CONTROLLER
// ========================================
//

int attenuator = 1;

//Slowly fade in or out LED attenuator in accordance with power state (called from Simple Timer every 10 ms)
void ledAttenuator() {

  if (debug) {
    //Serial.print(powerState());
  }

  // if S3 State (Standby Mode) is active, fade in and show "breathing" LED
  if (powerState() == 3) {
    attenuator = attenuator - 2;
  }
  // else fade LED fully OFF
  else {
    attenuator = attenuator + 2;
  }
  // Limit attenuator range
  attenuator = constrain(attenuator, 1, 255);
}

// Compute sinus wave for "breathing" sleep LED (Range 13 to 255)
void ledSinus() {
  float breathCycle = millis() / 4500.0; // 4500.0 means one sinus cycle per 4.5s
  float brightness = ( 134.0 + 121.0 * sin ( breathCycle * 2.0 * PI ) ) / attenuator;
  analogWrite(SLEEP_LED_POUT, brightness);
}

//
// ========================================
// CCFL BACKLIGHT INVERTER
// ========================================
//

int inverterBrightness;
boolean inverterEnabled;

//----Brightness

// This function increases the current brightness
void increaseBrightness () {
  adjustInverter(inverterBrightness = inverterBrightness + 10);
}

// This function decreases the current brightness
void decreaseBrightness () {
  adjustInverter(inverterBrightness = inverterBrightness - 10);
}

// This function limits and maps the inverter brightness and writes the value to the PWM output
// (called directly from Slider App (BWxx) or the Inc / Dec functions above)
void adjustInverter(int brightLevel) {
  // Limit brightness range
  brightLevel = constrain (brightLevel, 0, 100); // limit brightness range
  inverterBrightness = brightLevel; // Report brightness back to slider App!!

  // Map brightness 0 - 100% equals to 140 - 0 PWM (0 PWM = full brightness!!)
  brightLevel = map (brightLevel, 0, 100, 140, 0);
  
  if (inverterEnabled) {
    analogWrite(INVERTER_POUT, brightLevel);
  }
  else {
    analogWrite(INVERTER_POUT, 255); // 255 = Backlights OFF
  }
}

//----Inverter

// This function initilasises the inverter
void setupInverter() {
  inverterEnabled = true;
  adjustInverter(70); // brightness after SMC reset in %
}

// This function enables the Inverter and then writes the current brightness level to the PWM output
void enableInverter() {
  inverterEnabled = true;
  adjustInverter(inverterBrightness);
}

// This function disables the Inverter and then writes the current brightness level to the PWM output
void disableInverter() {
  inverterEnabled = false;
  adjustInverter(inverterBrightness);
}

//----Serial interface

// This function processes the inverter related serial commands
void processCommandBrightness(String subCmd, String extraCmd) {

  if (subCmd.equals("R")) {
    // Brightness Read (from the Arduino)
    Serial.println(inverterBrightness);

  } else if (subCmd.equals("W")) {
    // Brightness Write value BWnnn - nnn is the value
    adjustInverter( extraCmd.toInt() );

  } else if (subCmd.equals("A")) {
    // Activate Inverter
    enableInverter();

  } else if (subCmd.equals("D")) {
    // Deactivate Inverter
    disableInverter();

  } else if (subCmd.equals("+")) {
    // increase brightness
    increaseBrightness();

  } else if (subCmd.equals("-")) {
    // decrease brightness
    decreaseBrightness();

  } else {
    Serial.println(F("Brightness Command Unknown: BR (read), BW (write), BA (activate), BD (deactivate), B+, B- (change)"));
  }
}

//
// ========================================
// DRIVE DOOR SERVO
// ========================================
//

void servoFunction() {
  int pos;  //variable holds servo position
  int servoSpeed = 3; // 1 - 6
  int switchVal = digitalRead(CASE_POWER_SWITCH_PIN);  //Read switch

  if (switchVal == HIGH) {
    //pos = 0;
    pos = doorServo.read() - servoSpeed;

  } else {
    //pos = 180;
    pos = doorServo.read() + servoSpeed;

  }
  pos = constrain(pos, 0, 180);  //Limit servo position
  doorServo.write(pos);              // tell servo to go to position in variable 'pos'
}




//
// =====================
// USB SERIAL CONTROLLER
// =====================
//

// Some ASCII Constancts
const char CR = 13;
const char LF = 10;
const char TAB = '\t';
const char COMMA = ',';
const char EQUALS = '=';

const byte SERIAL_BAUD_RATE = 9600;

/**
 * PROCESS INPUT
 */
void loopSerialCommandProcessor() {

  static char serData[32]; // Buffer
  static int count = 0;
  static boolean initialised = false;

  if (! initialised) {
    // fist tie into this method setup
    Serial.begin(SERIAL_BAUD_RATE);
    initialised = true;
  }

  // if serial port not ready (Leonardo only).
  if (!Serial) return; // cant do anything

  while (Serial.available() > 0) {

    // read a byte and add to the buffer
    char inByte = Serial.read();

    // if th read failed
    if (inByte == -1) {
      break;

      // if EOL (End Of Line)
    } else if ( inByte == CR || inByte == LF ) {

      // terminate the buffer
      serData[count++] = 0;

      //Convert The array to a String
      String command = String(serData);

      if (command.length() > 0 ) {

        // process the command
        processCommand(command);
      }

      // clear the command buffer
      count = 0;

      // protect against overrun of buffer
    } else if (count < 30) {
      serData[count++] = inByte;
    }
  }
}

/**
 * Main function to process a command from terminal
*/
void processCommand(String cmd) {

  String firstCmd = cmd.substring(0, 1);
  String secondCmd = cmd.substring(1, 2);
  String extraCmd = cmd.substring(2);

  if (firstCmd.equals("B")) {
    processCommandBrightness(secondCmd, extraCmd);

  } /*else if (firstCmd.equals("C")) {
     processCommandCapacitive(secondCmd,extraCmd);

   } else if (firstCmd.equals("E")) {
     processCommandEeprom(secondCmd,extraCmd);

   } else if (firstCmd.equals("F")) {
     processCommandFan(secondCmd,extraCmd);

   } else if (firstCmd.equals("S")) {
     processCommandSystem(secondCmd,extraCmd);

  //  } else if (firstCmd.equals("P")) {
  //    processCommandPSU(secondCmd,extraCmd);

   } else {
     Serial.println(F("Command Unknown: B (brighness), C (capacitive), E (eeprom), F (fans), P (psu), S (system)"));
   }

   // todo command to test the chime.

   // todo command to flash the LED

   // todo read current input temp sensors*/
}

void processCommandSystem(String subCmd, String extraCmd) {

  if (subCmd.equals("R")) {
    /*   // System Ram Available in bytes. Leonardo has 2.5Kbyes
       Serial.print(F("FreeRAM "));
       Serial.print(freeRam());
       Serial.println(F(" of 2560 bytes"));

     } else if (subCmd.equals("U")) {
       // System UpTime in seconds
       Serial.print(F("UpTime "));
       Serial.print(millis()/1000);
       Serial.print(F("."));
       Serial.print(millis()%1000);
       Serial.println(F(" seconds"));

     } else if (subCmd.equals("T")) {
       // System Timers in-use
       Serial.print(F("Timers In Use "));
       Serial.println(timer.getNumTimers());
       Serial.print(F("Timers Unused "));
       Serial.println(timer.getNumAvailableTimers());

     } else if (subCmd.equals("I")) {
       // Read System Inputs
       Serial.print(F("CapDNSensor\t"));
       Serial.println(getDnSensorReading());
       Serial.print(F("CapUPSensor\t"));
       Serial.println(getUpSensorReading());
       Serial.print(F("MainTemp\t"));
       Serial.println(getTemp());
       Serial.print(F("AmbientTemp\t"));
       Serial.println(getAmbientTemp());
       Serial.print(F("PowerState\t"));
       Serial.println(powerState());
       Serial.print(F("PowerButton\t"));
       Serial.println(powerButtonDetection());
    */
  } else {
    Serial.println(F("System Command Unknown: R (ram), U (uptime), T (timers), I (input)"));
  }
}

//
// =========================================
// SMOOTHING FUNCTION
// =========================================
//

int smoothInt(int raw, int readings[], int arrayLength, int& pointer, long& total) {

  // declare function internal variables!
  int average;

  // subtract the last reading:
  total = total - readings[pointer];
  // read from the sensor:
  readings[pointer] = raw;
  // add the reading to the total:
  total = total + readings[pointer];
  // advance to the next position in the array:
  pointer = pointer + 1;

  // if we're at the end of the array...
  if (pointer >= arrayLength)
    // ...wrap around to the beginning:
    pointer = 0;

  // calculate the average:
  average = total / arrayLength;

  /*if (debug) {
    Serial.print("Raw = ");
    Serial.println(raw);

    Serial.print("Readings = ");
    Serial.println(readings[pointer]);

    Serial.print("Pointer = ");
    Serial.println(pointer);

    Serial.print("Array Length = ");
    Serial.println(arrayLength);

    Serial.print("Total = ");
    Serial.println(total);
  }*/

  // return the smoothed value
  return average;
}
