//
// -----------------------
// PWM FREQUENCY PRESCALER
// -----------------------
// Written by Kiwisincebirth 2014
// Revised by MacTester57 to allow correct PWM frequencies (confirmed with oscilloscope). January 2015
// - prescale variable: replaced uint8_t with uint16_t data type (fixes bug, which did not allow frequencies < 492Hz)
// - mode variable: replaced hex format with binary to make it more readable, if compared with bit tables in the 32U4 manual
// - added comments

/**
 * Arduino Leonardo AtMega 32u4 specific
 * Sets the Prescaler (Divisor) for a given PWM pin. The resulting frequency
 * is equal to the base frequency divided by the given divisor:
 *   - Base frequencies:
 *      o The base frequency for pins 3, and 11 is 64,500 Hz.
 *      o The base frequency for pins 5, 9, and 10 is 31,250 Hz.
 *      o The base frequency for pins 6 and 13 is 125,000 Hz.
 *   - Divisors:
 *      o The divisors available on pins 3, 5, 9, 10 and 11 are: 1, 8, 64, 256, and 1024.
 *      o The divisors available on pins 6 and 13 are all powers of two between 1 and 16384
 *
 * PWM frequencies are tied together in pairs of pins. If one in a
 * pair is changed, the other is also changed to match:
 *   - Pins 3 and 11 are paired on timer0 (8bit)
 *   - Pins 9 and 10 are paired on timer1 (16bit)
 *   - Pins 6 and 13 are paired on timer4 (10bit)
 *   - Pins 5 is exclusivly on timer3 (16bit)
 *
 * Note: Pins 3 and 11 operate on Timer 0 changes on this pins will
 * affect the user of the main time millis() functions
 */
void setPWMPrescaler(uint8_t pin, uint16_t prescale) {
  if (pin == 3 || pin == 11) {  // Pin3 tested, OK, but millis() is affected!! (see 32U4 manual page 105-106)
    byte mode;
    switch (prescale) {
      case 1: mode = 0b001; break; //CK (31.5kHz)
      case 8: mode = 0b010; break; //CK/8
      case 64: mode = 0b011; break; //CK/64
      case 256: mode = 0b100; break; //CK/256 (123Hz)
      case 1024: mode = 0b101; break; //CK/1024
      default: return;
    }
    TCCR0B = TCCR0B & 0b11111000 | mode;  // timer0 (8bit)
  }

  if (pin == 5 || pin == 9 || pin == 10) { //(see 32U4 manual page 133)
    byte mode;
    switch (prescale) {
      case 1: mode = 0b001; break; //CK (31.5kHz)
      case 8: mode = 0b010; break; //CK/8
      case 64: mode = 0b011; break; //CK/64
      case 256: mode = 0b100; break; //CK/256 (123Hz)
      case 1024: mode = 0b101; break; //CK/1024
      default: return;
    }
    if (pin == 9 || pin == 10) {  // Pin 5, 9 & 10 tested, OK
      TCCR1B = TCCR1B & 0b11111000 | mode;  // timer1 (16bit)
    } else {
      TCCR3B = TCCR3B & 0b11111000 | mode;  // timer3 (16bit)
    }
  }

  if (pin == 6 || pin == 13) {  // Pin 6 tested, with oscilloscope! (see 32U4 manual pages 165-166)
    byte mode;
    switch (prescale) {
      case 1: mode = 0b0001; break; //CK (31.5kHz)
      case 2: mode = 0b0010; break; //CK/2
      case 4: mode = 0b0011; break; //CK/4
      case 8: mode = 0b0100; break; //CK/8
      case 16: mode = 0b0101; break; //CK/16
      case 32: mode = 0b0110; break; //CK/32
      case 64: mode = 0b0111; break; //CK/64 (492Hz)
      case 128: mode = 0b1000; break; //CK/128
      case 256: mode = 0b1001; break; //CK/256 (123Hz)
      case 512: mode = 0b1010; break; //CK/512
      case 1024: mode = 0b1011; break; //CK/1024
      case 2048: mode = 0b1100; break; //CK/2048
      case 4096: mode = 0b1101; break; //CK/4096
      case 8192: mode = 0b1110; break; //CK/8192
      case 16384: mode = 0b1111; break; //CK/16384
      default: return;
    }
    TCCR4B = TCCR4B & 0b11110000 | mode;  // timer4 (10bit)
  }
}
